# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
from copy import copy
from itertools import product
import pytest
import networkx as nx


from hnccorr.segmentation import HncParametricWrapper, Segmentation


class TestHNC:
    def test_hnc(self):
        G = nx.DiGraph()
        G.add_nodes_from((i,) for i in range(7))

        h = HncParametricWrapper(0, 2)

        segmentations = h.solve(G, {(2,), (3,), (4,)}, {(1,), (5,)})

        assert len(segmentations) == 1
        assert segmentations[0].selection == {(0,), (2,), (3,), (4,), (6,)}
        assert segmentations[0].weight == pytest.approx(2.0)


class TestSegmentation:
    def test_segmentation_weight(self):
        assert Segmentation({0, 1}, 0.5).weight == 0.5

    def test_segmentation_selection(self):
        assert Segmentation({(0, 1)}, 0.5).selection == {(0, 1)}

    def test_segmentation_equality_wrong_class(self):
        class FakeSegmentation:
            pass

        assert Segmentation({(0, 1)}, 0.5) != FakeSegmentation()

    def test_segmentation_clean_fill_hole(self):
        selection = {(0, 0), (1, 0), (2, 0), (0, 1), (2, 1), (0, 2), (1, 2), (2, 2)}
        weight = 1
        original = Segmentation(selection, weight)

        new = original.clean(set(), (5, 5))
        assert new == Segmentation(
            {(0, 0), (1, 0), (2, 0), (0, 1), (1, 1), (2, 1), (0, 2), (1, 2), (2, 2)},
            weight,
        )
        assert original == Segmentation(selection, weight)

    def test_segmentation_clean_select_seed_component(self):
        selection = {(0, 0), (3, 3)}
        weight = 1
        original = Segmentation(selection, weight)

        new = original.clean({(3, 3)}, (5, 5))
        assert new == Segmentation({(3, 3)}, weight)
        assert original == Segmentation(selection, weight)



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
"""Components for calcium-imaging movies in HNCcorr."""

import os
import math
import numpy as np
from PIL import Image
from PIL.TiffTags import TAGS

from hnccorr.utils import (
    add_offset_to_coordinate,
    add_offset_set_coordinates,
    add_time_index,
    generate_pixels,
    list_images,
)


class Movie:
    """Calcium imaging movie class.

    Data is stored in an in-memory numpy array. Class supports both 2- and 3-
    dimensional movies.

    Attributes:
        name(str): Name of the experiment.
        _data (np.array): Fluorescence data. Array has size T x N1 x N2. T is
            the number of frame (num_frames), N1 and N2 are the number of
            pixels in the first and second dimension respectively.
        _data_size (tuple): Size of array _data.
    """

    def __init__(self, name, data):
        self.name = name
        self._data = data
        self.data_size = data.shape

    @classmethod
    def from_tiff_images(cls, name, image_dir, num_images, memmap=False, subsample=10):
        """Loads tiff images into a numpy array.

        Data is assumed to be stored in 16-bit unsigned integers. Frame numbers are
        assumed to be padded with zeros: 00000, 00001, 00002, etc. This is required
        such that Python sorts the images correctly. Frame numbers can start from 0, 1,
        or any other number. Files must have the extension ``.tiff``.

        If memmap is True, the data is not loaded into memory bot a memory mapped file
        on disk is used. The file is named ``$name.npy`` and is placed in the
        `image_dir` folder.

        Args:
            name (str): Movie name.
            image_dir (str): Path of image folder.
            num_images (int): Number of images in the folder.
            memmap (bool): If True, a memory-mapped file is used. (*Default: False*)
            subsample (int): Number of frames to average into a single frame.

        Returns:
            Movie: Movie created from image files.
        """

        images, data_size = cls._get_tiff_images_and_size(image_dir, num_images)

        subsampler = Subsampler(data_size, subsample)

        if memmap:
            memmap_filename = os.path.join(image_dir, name + ".npy")
            data = np.memmap(
                memmap_filename,
                dtype=np.float32,
                mode="w+",
                shape=subsampler.output_shape,
            )
        else:
            data = np.zeros(subsampler.output_shape, np.float32)

        cls._read_images(images, data, subsampler)

        return cls(name, data)

    @staticmethod
    def _get_tiff_images_and_size(image_dir, num_images):
        """ Provides a sorted list of images and computes the required array size.

        Data is assumed to be stored in 16-bit unsigned integers. Frame numbers are
        assumed to be padded with zeros: 00000, 00001, 00002, etc. This is required
        such that Python sorts the images correctly. Frame numbers can start from 0, 1,
        or any other number. Files must have the extension ``.tiff``.

        Args:
            image_dir (str): Path of image folder.
            num_images (int): Number of images in the folder.

        Returns:
            tuple[List[Str], tuple]: Tuple of the list of images and the array size.
        """
        images = list_images(image_dir)

        assert len(images) == num_images

        # read image meta data
        first_image = images[0]
        with Image.open(first_image) as image:
            meta = {TAGS[key]: image.tag[key] for key in image.tag}

        # set size of data
        data_size = (len(images), meta["ImageLength"][0], meta["ImageWidth"][0])

        return images, data_size

    @staticmethod
    def _read_images(images, output_array, subsampler):
        """ Loads images and copies them into the provided array.

        Args:
            images (list[Str]): Sorted list image paths.
            output_array (np.array like): T x N_1 x N_2 array-like object into which
                images should be loaded. T must equal the number of images in `images`.
                Each image should be of size N_1 x N_2.
            subsampler

        Returns:
            np.array like: The input array `array`.

        """
        for filename in images:
            if subsampler.buffer_full:
                output_array[
                    slice(*subsampler.buffer_indices), :, :
                ] = subsampler.buffer
                subsampler.advance_buffer()

            with Image.open(filename) as image:
                subsampler.add_frame(np.array(image))

        output_array[slice(*subsampler.buffer_indices), :, :] = subsampler.buffer

        return output_array

    def __getitem__(self, key):
        """Provides direct access to the movie data.

        Movie is stored in array with shape (T, N_1, N_2, ...), where T is the number
        of frames in the movie. N_1, N_2, ... are the number of pixels in the first
        dimension, second dimension, etc.

        Args:
            key (tuple): Valid index for a numpy array.

        Returns:
            np.array
        """
        return self._data.__getitem__(key).astype(np.float64)

    def is_valid_pixel_coordinate(self, coordinate):
        """Checks if coordinate is a coordinate for a pixel in the movie."""
        if self.num_dimensions != len(coordinate):
            return False

        zero_tuple = (0,) * self.num_dimensions
        for i, lower, upper in zip(coordinate, zero_tuple, self.pixel_shape):
            if not lower <= i < upper:
                return False
        return True

    @property
    def num_frames(self):
        """Number of frames in the movie."""
        return self.data_size[0]

    @property
    def pixel_shape(self):
        """Resolution of the movie in pixels."""
        return self.data_size[1:]

    @property
    def num_pixels(self):
        """Number of pixels in the movie."""
        return np.product(self.data_size[1:])

    @property
    def num_dimensions(self):
        """Dimension of the movie (excludes time dimension)."""
        return len(self.data_size[1:])

    def extract_valid_pixels(self, pixels):
        """Returns subset of pixels that are valid coordinates for the movie."""
        return {pixel for pixel in pixels if self.is_valid_pixel_coordinate(pixel)}


class Patch:
    """Square subregion of Movie.

    Patch limits the data used for the segmentation of a potential cell. Given a center
    seed pixel, Patch defines a square subregion centered on the seed pixel with width
    patch_size. If the square extends outside the movie boundaries, then the subregion
    is shifted such that it stays within the movie boundaries.

    The patch also provides an alternative coordinate system with respect to the top
    left pixel of the patch. This pixel is the zero coordinate for the patch coordinate
    system. The coordinate offset is the coordinate of the top left pixel in the movie
    coordinate system.

    Attributes:
        _center_seed (tuple): Seed pixel that marks the potential cell. The pixel is
            represented as a tuple of coordinates. The coordinates are relative to the
            movie. The top left pixel of the movie represents zero.
        _coordinate_offset (tuple): Movie coordinates of the pixel that represents the
            zero coordinate in the Patch object. Similar to the Movie, pixels in the
            Patch are indexed from the top left corner.
        _data (np.array): Subset of the Movie data. Only data for the patch is stored.
        _movie (Movie): Movie for which the Patch object is a subregion.
        _num_dimensions (int): Dimension of the patch. It matches the dimension of the
            movie.
        _patch_size (int): length of the patch in each dimension. Must be an odd number.
    """

    def __init__(self, movie, center_seed, patch_size):
        """Initializes Patch object."""
        if patch_size % 2 == 0:
            raise ValueError("patch_size (%d) should be an odd number.")

        self._num_dimensions = movie.num_dimensions
        self._center_seed = center_seed
        self._patch_size = patch_size
        self._movie = movie
        self._coordinate_offset = self._compute_coordinate_offset()
        self._data = self._movie[self._movie_indices()]

    @property
    def num_frames(self):
        """Number of frames in the Movie."""
        return self._movie.num_frames

    @property
    def pixel_shape(self):
        """Shape of the patch in pixels. Does not not included the time dimension."""
        return (self._patch_size,) * self._num_dimensions

    def _compute_coordinate_offset(self):
        """Computes the coordinate offset of the patch.

        Confirms that the patch falls within the movie boundaries and shifts the patch
        if necessary. The center seed pixel may not be in the center of the patch if a
        shift is necessary.
        """
        half_width = int((self._patch_size - 1) / 2)

        topleft_coordinates = add_offset_to_coordinate(
            self._center_seed, (-half_width,) * self._num_dimensions
        )
        # shift left such that top left corner exists
        topleft_coordinates = list(max(x, 0) for x in topleft_coordinates)

        # bottomright corners (python-style index so not included)
        bottomright_coordinates = add_offset_to_coordinate(
            topleft_coordinates, (self._patch_size,) * self._num_dimensions
        )
        # shift right such that bottom right corner exists
        bottomright_coordinates = list(
            min(x, max_value)
            for x, max_value in zip(bottomright_coordinates, self._movie.pixel_shape)
        )

        topleft_coordinates = add_offset_to_coordinate(
            bottomright_coordinates, (-self._patch_size,) * self._num_dimensions
        )

        return topleft_coordinates

    def _movie_indices(self):
        """Computes the indices of the movie that correspond to the patch.

        For a patch with top left pixel (5, 5) and bottom right pixel (9, 9), this
        method returns ``(:, 5:10, 5:10)`` which can be used to acccess the data
        corresponding to the patch in the movie.
        """
        bottomright_coordinates = add_offset_to_coordinate(
            self._coordinate_offset, (self._patch_size,) * self._num_dimensions
        )

        # pixel indices
        idx = []
        for start, stop in zip(self._coordinate_offset, bottomright_coordinates):
            idx.append(slice(start, stop))
        return add_time_index(tuple(idx))

    def to_movie_coordinate(self, patch_coordinate):
        """Converts a movie coordinate into a patch coordinate.

        Args:
            patch_coordinate (tuple): Coordinates of a pixel in patch coordinate system.

        Returns:
            tuple: Coordinate of pixel in movie coordinate system.
        """
        return add_offset_to_coordinate(patch_coordinate, self._coordinate_offset)

    def to_patch_coordinate(self, movie_coordinate):
        """Converts a movie coordinate into a patch coordinate.

        Args:
            movie_coordinate (tuple): Coordinates of a pixel in movie coordinate system.

        Returns:
            tuple: Coordinate of pixel in patch coordinate system.
        """
        return add_offset_to_coordinate(
            movie_coordinate, [-x for x in self._coordinate_offset]
        )

    def enumerate_pixels(self):
        """Returns the movie coordinates of the pixels in the patch."""
        return add_offset_set_coordinates(
            generate_pixels(self.pixel_shape), self._coordinate_offset
        )

    def __getitem__(self, key):
        """Access data for pixels in the patch. Indexed in patch coordinates."""
        return self._data[key]


class Subsampler:
    """Subsampler for averaging frames.

    Averages `subsample_frequency` into a single frame. Stores averaged frames in a
    buffer and writes buffer to an output array.

    Attributes:
        _buffer (np.array): (b, N_1, N_2) array where the frame averages are compiled.
        _buffer_frame_count: (b, ) array with the number of frames used in each
            averaged frame.
        _buffer_size (int): Number of averaged frames to store in buffer. Short: b.
            Default is 10.
        _buffer_start_index (int): Index of averaged movie corresponding with first
            frame in the buffer.
        _current_index (int): Index of current frame in buffer.
        _movie_shape (int): Shape of input movie.
        _num_effective_frames (int): Number of frames in the averaged movie.
        _subsample_frequency (int): Number of frames to average into a single frame.
    """

    def __init__(self, movie_shape, subsample_frequency, buffer_size=10):
        """Initializes a subsampler object."""
        self._movie_shape = movie_shape
        self._subsample_frequency = subsample_frequency
        self._num_effective_frames = int(
            math.ceil(self._movie_shape[0] / float(self._subsample_frequency))
        )
        self._buffer_size = min(buffer_size, self._num_effective_frames)
        self._current_index = 0

        self._buffer = np.zeros(
            (self._buffer_size, *self._movie_shape[1:]), dtype=np.float32
        )

        self._buffer_frame_count = np.zeros((self._buffer_size,))
        self._buffer_start_index = 0

    @property
    def output_shape(self):
        """Shape of average movie array."""
        return (self._num_effective_frames, *self._movie_shape[1:])

    @property
    def buffer(self):
        """Provides access to data in buffer. Corrects last buffer for movie length."""
        # x % 10 takes values in 0...9 whereas x - 1 % 10 + 1 takes values 1, .. 10
        max_index = ((self.buffer_indices[1] - 1) % self._buffer_size) + 1
        return self._buffer[:max_index, :, :]

    @property
    def buffer_full(self):
        """True if buffer is full."""
        return self._current_index >= self._buffer_size

    @property
    def buffer_indices(self):
        """Indices in average movie corresponding to current buffer"""
        return (
            self._buffer_start_index,
            min(
                self._buffer_start_index + self._buffer_size, self._num_effective_frames
            ),
        )

    def add_frame(self, frame):
        """ Adds frame to average.

        Frames should be provided in order of appearance in the movie.

        Args:
            frame (np.array): (N_1, N_2) array with pixel intensities.

        Returns:
            None

        Raises:
            ValueError: If buffer is full.
        """
        if self.buffer_full:
            raise ValueError("Buffer is full. Cannot add current frame.")

        current_num_frames = self._buffer_frame_count[self._current_index]
        new_num_frames = float(current_num_frames + 1)
        self._buffer[self._current_index, :, :] = (
            current_num_frames
            / new_num_frames
            * self._buffer[self._current_index, :, :]
            + frame / new_num_frames
        )
        self._buffer_frame_count[self._current_index] += 1

        # advance to next frame in buffer if current frame is full
        if self._buffer_frame_count[self._current_index] == self._subsample_frequency:
            self._current_index += 1

    def advance_buffer(self):
        """Empties buffer and advances the buffer indices for new frames"""
        self._buffer_frame_count[:] = 0
        self._buffer[:] = 0
        self._current_index = 0
        self._buffer_start_index += self._buffer_size



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
import pytest
from copy import deepcopy

from hnccorr.base import HNCcorr, Candidate, HNCcorrConfig, DEFAULT_CONFIG
from hnccorr.segmentation import Segmentation


@pytest.fixture
def config():
    return HNCcorrConfig(patch_size=31, negative_seed_circle_radius=10)


@pytest.fixture
def candidate(H):
    return Candidate(H.seeder.return_val, H)


@pytest.fixture
def mock_candidate_class(mocker):
    return mocker.patch("hnccorr.base.Candidate", autospec=True)


@pytest.fixture
def mock_seeder(mocker, dummy):
    return mocker.patch("hnccorr.seeds.LocalCorrelationSeeder", autospec=True)(
        dummy, dummy, dummy, dummy
    )


@pytest.fixture
def mock_pos_seed_selector(mocker, dummy):
    pos_seed_selector = mocker.patch(
        "hnccorr.seeds.PositiveSeedSelector", autospec=True
    )(dummy)

    return pos_seed_selector


@pytest.fixture
def mock_neg_seed_selector(mocker, dummy):
    neg_seed_selector = mocker.patch(
        "hnccorr.seeds.NegativeSeedSelector", autospec=True
    )(dummy, dummy)
    return neg_seed_selector


@pytest.fixture
def mock_segmentor(mocker, dummy):
    segmentor = mocker.patch(
        "hnccorr.segmentation.HncParametricWrapper", autospec=True
    )(dummy, dummy)
    return segmentor


@pytest.fixture
def mock_postprocessor(mocker, dummy):
    postprocessor = mocker.patch(
        "hnccorr.postprocessor.SizePostprocessor", autospec=True
    )(dummy, dummy, dummy)
    return postprocessor


@pytest.fixture
def mock_graph_constructor(mocker, dummy):
    graph_constructor = mocker.patch("hnccorr.graph.GraphConstructor", autospec=True)(
        dummy, dummy
    )
    return graph_constructor


@pytest.fixture
def mock_patch_class(mocker, dummy):
    patch_class = mocker.patch("hnccorr.movie.Patch", autospec=True)
    return patch_class


@pytest.fixture
def mock_embedding_class(mocker, dummy):
    embedding_class = mocker.patch("hnccorr.graph.CorrelationEmbedding", autospec=True)
    return embedding_class


@pytest.fixture
def mock_segmentation_class(mocker):
    return mocker.patch("hnccorr.segmentation.Segmentation", autospec=True)


@pytest.fixture
def H(
    dummy,
    seeder_fixed_val,
    postprocessor_select_first,
    segmentor_simple_segmentation,
    mock_candidate_class,
):
    return HNCcorr(
        seeder_fixed_val,
        "postprocessor",
        segmentor_simple_segmentation,
        "pos_seed_selector",
        "neg_seed_selector",
        "graph_constructor",
        mock_candidate_class,
        "patch",
        "embedding",
        "patch_size",
    )


@pytest.fixture
def hnccorr(
    dummy,
    mock_movie,
    mock_postprocessor,
    mock_segmentor,
    mock_pos_seed_selector,
    mock_neg_seed_selector,
    mock_graph_constructor,
    mock_patch_class,
    mock_embedding_class,
):
    H = HNCcorr(
        dummy,
        mock_postprocessor,
        mock_segmentor,
        mock_pos_seed_selector,
        mock_neg_seed_selector,
        mock_graph_constructor,
        Candidate,
        mock_patch_class,
        mock_embedding_class,
        "patch_size",
    )

    H.movie = mock_movie
    return H


class TestCandidate:
    def test_candidate_segment(
        self,
        hnccorr,
        dummy,
        mock_movie,
        mock_postprocessor,
        mock_segmentor,
        mock_pos_seed_selector,
        mock_neg_seed_selector,
        mock_graph_constructor,
        mock_patch_class,
        mock_embedding_class,
        mock_segmentation_class,
    ):
        mock_pos_seed_selector.select.return_value = "positive_seed"
        mock_neg_seed_selector.select.return_value = "negative_seed"
        mock_embedding_class.return_value = "embedding_class"
        mock_postprocessor.select.return_value = "best segmentation"
        mock_graph_constructor.construct.return_value = "graph"
        mock_patch_class.return_value = "patch"
        mock_segmentation_class.return_value.clean.return_value = "clean"

        center_seed = 1
        mock_segmentor.solve.return_value = [
            mock_segmentation_class(dummy, dummy),
            mock_segmentation_class(dummy, dummy),
        ]

        candidate = Candidate(center_seed, hnccorr)
        assert candidate.segment() == mock_postprocessor.select.return_value
        mock_pos_seed_selector.select.assert_called_once_with(center_seed, mock_movie)
        mock_neg_seed_selector.select.assert_called_once_with(center_seed, mock_movie)
        mock_patch_class.assert_called_once_with(mock_movie, center_seed, "patch_size")
        mock_embedding_class.assert_called_once_with(mock_patch_class.return_value)
        mock_graph_constructor.construct.assert_called_once_with(
            mock_patch_class.return_value, mock_embedding_class.return_value
        )
        mock_segmentor.solve.assert_called_once_with(
            mock_graph_constructor.construct.return_value,
            mock_pos_seed_selector.select.return_value,
            mock_neg_seed_selector.select.return_value,
        )
        mock_postprocessor.select.assert_called_once_with(["clean", "clean"])

    def test_candidate_equality(self):
        assert Candidate(1, "a") == Candidate(1, "a")
        assert Candidate(1, "a") != Candidate(2, "a")
        assert Candidate(1, "a") != Candidate(1, "b")

    def test_candidate_equality_wrong_class(self):
        class FakeCandidate:
            pass

        assert Candidate(1, "a") != FakeCandidate()

    def test_candidate_segmentations(self, hnccorr, mock_segmentor):
        c = Candidate(1, hnccorr)

        assert c.segmentations is None
        c.segment()
        assert c.segmentations == mock_segmentor.solve.return_value

    def test_candidate_center_seed(self):
        assert Candidate(1, "a").center_seed == 1

    def test_candidate_clean_segmentations(
        self, dummy, mock_segmentor, mock_segmentation_class, hnccorr
    ):
        mock_segmentor.solve.return_value = [
            mock_segmentation_class(dummy, dummy),
            mock_segmentation_class(dummy, dummy),
        ]
        mock_segmentation_class.return_value.clean.return_value = "clean"

        c = Candidate(1, hnccorr)
        assert c.clean_segmentations is None
        c.segment()
        assert c.clean_segmentations == ["clean", "clean"]

    def test_candidate_best_segmentations(self, hnccorr, mock_postprocessor):
        c = Candidate(1, hnccorr)
        assert c.best_segmentation is None
        c.segment()
        assert c.best_segmentation == mock_postprocessor.select.return_value


class TestHNCcorr:
    def test_hnccorr_seeder(self, H, seeder_fixed_val):
        assert H.seeder == seeder_fixed_val

    def test_hnccorr_from_config(self):
        assert isinstance(HNCcorr.from_config(HNCcorrConfig()), HNCcorr)

    def test_hnccorr_from_config(self):
        assert isinstance(HNCcorr.from_config(), HNCcorr)

    def test_hnccorr_from_non_default_config(self):
        assert isinstance(HNCcorr.from_config(HNCcorrConfig()), HNCcorr)

    def test_hnccorr_movie(self, H, MM, seeder_fixed_val):
        seeder_fixed_val.called = True  # deactive seeder

        assert H.movie is None
        H.segment(MM)
        assert H.movie == MM

    def test_hnccorr_patch_size(self, H):
        assert H.patch_size == "patch_size"

    def test_hnccorr_seeder(self, H, seeder_fixed_val):
        assert H.seeder == seeder_fixed_val

    def test_hnccorr_patch_class(self, H):
        assert H.patch_class == "patch"

    def test_hnccorr_embedding_class(self, H):
        assert H.embedding_class == "embedding"

    def test_hnccorr_segment_calls_candidate_segment(self, H, MM, mock_candidate_class):
        H.segment(MM)
        assert mock_candidate_class.return_value.segment.call_count > 0

    def test_hnccorr_positive_seed_selector(self, H):
        assert H.positive_seed_selector == "pos_seed_selector"

    def test_hnccorr_negative_seed_selector(self, H):
        assert H.negative_seed_selector == "neg_seed_selector"

    def test_hnccorr_graph_constructor(self, H):
        assert H.graph_constructor == "graph_constructor"

    def test_hnccorr_seeder(self, H, postprocessor_select_first):
        assert H.postprocessor == postprocessor_select_first

    def test_hnccorr_seeder(self, H, segmentor_simple_segmentation):
        assert H.segmentor == segmentor_simple_segmentation

    def test_hnccorr_segmentations_initialization(self, H):
        assert H.segmentations == []

    def test_hnccorr_candidates_initialization(self, H):
        assert H.candidates == []

    def test_hnccorr_candidates_after_segment(self, H, MM, mock_candidate_class):
        base_segmentation = Segmentation(set(), "segment")
        mock_candidate_class.return_value.segment.return_value = base_segmentation

        assert H.candidates == []
        H.segment(MM)
        assert len(H.candidates) == 1
        assert H.candidates[0].segment() == base_segmentation

        H.segment(MM)
        assert len(H.candidates) == 1
        assert H.candidates[0].segment() == base_segmentation

    def test_hnccorr_segmentations_after_segment(self, H, MM, mock_candidate_class):
        base_segmentation = Segmentation(set(), "segment")
        mock_candidate_class.return_value.segment.return_value = base_segmentation

        H.segment(MM)
        assert H.segmentations == [base_segmentation]

        H.segment(MM)
        assert H.segmentations == [base_segmentation]

    def test_hnccorr_segment_none_is_not_added_to_segmentations(
        self, MM, H, mock_candidate_class
    ):
        mock_candidate_class.return_value.segment.return_value = None

        H.segment(MM)

        assert H.segmentations == []

    def test_hnccor_exclude_previously_segmented_pixels(
        self, mock_seeder, dummy, mock_candidate_class
    ):
        mock_seeder.next.side_effect = ("seed1", None)
        mock_candidate_class.return_value.segment.return_value = Segmentation(
            {"pixel"}, 1
        )
        H = HNCcorr(
            mock_seeder,
            "postprocessor",
            "mock_segmentor",
            "pos_seed_selector",
            "neg_seed_selector",
            "graph_constructor",
            mock_candidate_class,
            "patch",
            "embedding",
            "patch_size",
        )

        H.segment(dummy)

        mock_seeder.exclude_pixels.assert_called_once_with({"pixel"})

    def test_segmentations_to_list(self, H, dummy, mock_segmentation_class):
        # selections are stored as lists to fix order.
        ms1 = mock_segmentation_class(dummy, dummy)
        ms1.selection = [(0, 1), (1, 0)]
        ms2 = deepcopy(ms1)
        ms2.selection = [(8, 3), (3, 10), (138, 23)]
        H.segmentations = [ms1, ms2]

        assert H.segmentations_to_list() == [
            {"coordinates": [(0, 1), (1, 0)]},
            {"coordinates": [(8, 3), (3, 10), (138, 23)]},
        ]


class TestHnccorrConfig:
    def test_config_attributes(self, config):
        assert config.patch_size == 31
        assert config.negative_seed_circle_radius == 10

    def test_invalid_parameter(self):
        with pytest.raises(ValueError):
            HNCcorrConfig(wrong_parameter=10)

    def test_config_add_wrong_class(self, config):
        class WrongClass:
            pass

        with pytest.raises(TypeError):
            config + WrongClass()

    def test_config_add_config(self, config):
        config2 = HNCcorrConfig(patch_size=21, positive_seed_radius=5)

        config3 = config + config2
        assert config.patch_size == 31
        assert config2.patch_size == 21
        assert config3.patch_size == 21

        with pytest.raises(AttributeError):
            config2.negative_seed_circle_radius
        assert config.negative_seed_circle_radius == 10
        assert config3.negative_seed_circle_radius == 10

        with pytest.raises(AttributeError):
            config.positive_seed_radius == 5
        assert config2.positive_seed_radius == 5
        assert config3.positive_seed_radius == 5

    def test_config_default_config(self):
        assert DEFAULT_CONFIG.seeder_mask_size == 3
        assert DEFAULT_CONFIG.seeder_exclusion_padding == 4
        assert DEFAULT_CONFIG.seeder_grid_size == 5
        assert DEFAULT_CONFIG.percentage_of_seeds == pytest.approx(0.4)
        assert DEFAULT_CONFIG.postprocessor_min_cell_size == 40
        assert DEFAULT_CONFIG.postprocessor_max_cell_size == 200
        assert DEFAULT_CONFIG.postprocessor_preferred_cell_size == 80
        assert DEFAULT_CONFIG.positive_seed_radius == 0
        assert DEFAULT_CONFIG.negative_seed_circle_radius == pytest.approx(10.0)
        assert DEFAULT_CONFIG.negative_seed_circle_count == 10
        assert DEFAULT_CONFIG.gaussian_similarity_alpha == pytest.approx(1.0)
        assert DEFAULT_CONFIG.sparse_computation_grid_distance == pytest.approx(
            1 / 35.0
        )
        assert DEFAULT_CONFIG.sparse_computation_dimension == 3
        assert DEFAULT_CONFIG.patch_size == 31



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
import pytest
import os
import numpy as np
from copy import copy, deepcopy
from pytest_mock import mocker

from hnccorr.movie import Movie, Patch, Subsampler

from conftest import TEST_DATA_DIR


@pytest.fixture
def movie_data():
    data = np.zeros((3, 5, 10), np.float32)
    data[0, :, :] = np.ones((5, 10))
    data[1, :, :] = np.ones((5, 10)) * 2
    data[2, :, :] = np.ones((5, 10)) * 3
    return data


@pytest.fixture
def M(movie_data):
    return Movie("Simple", movie_data)


@pytest.fixture
def simple_patch(MM):
    return Patch(MM, (9,), 7)


@pytest.fixture
def subsampler():
    return Subsampler((800, 10, 10), 10)


class TestMovie:
    def test_movie_from_tiff_images(self, movie_data):
        """
        Movie consists of three images of 5 x 10 pixels. All pixels in the first
        image have value 1, all pixels in the second image have value 2, and all
        pixels in the third image have value 3.
        """
        movie_from_tiff = Movie.from_tiff_images(
            "Simple",
            image_dir=str(os.path.join(TEST_DATA_DIR, "simple_movie")),
            num_images=3,
            subsample=1,
        )

        # compare data of movie from_tiff and direct initialization.
        np.testing.assert_allclose(movie_from_tiff[:], movie_data)

    def test_movie_from_tiff_images_with_subsampling(self, movie_data):
        """
        Movie consists of three images of 5 x 10 pixels. All pixels in the first
        image have value 1, all pixels in the second image have value 2, and all
        pixels in the third image have value 3.
        """
        movie_from_tiff = Movie.from_tiff_images(
            "Simple",
            image_dir=str(os.path.join(TEST_DATA_DIR, "simple_movie_long")),
            num_images=21,
            subsample=2,
        )

        # compare data of movie from_tiff and direct initialization.

        data = np.zeros((11, 5, 10), np.float32)
        data[0, :, :] = np.ones((5, 10)) * 1.5
        for i in range(1, 11):
            data[i, :, :] = np.ones((5, 10)) * 2
        np.testing.assert_allclose(movie_from_tiff[:], data)

    def test_movie_from_tiff_images_memmap(self, movie_data):
        """
        Movie consists of three images of 5 x 10 pixels. All pixels in the first
        image have value 1, all pixels in the second image have value 2, and all
        pixels in the third image have value 3.
        """
        movie_from_tiff = Movie.from_tiff_images(
            "Simple",
            image_dir=str(os.path.join(TEST_DATA_DIR, "simple_movie")),
            num_images=3,
            memmap=True,
            subsample=1,
        )

        # compare data of movie from_tiff and direct initialization.
        np.testing.assert_allclose(movie_from_tiff[:], movie_data)

    def test_movie_name(self, M):
        assert M.name == "Simple"

    def test_movie_num_frames(self, M):
        assert M.num_frames == 3

    def test_movie_num_pixels(self, M):
        assert M.num_pixels == 50

    def test_movie_data_size(self, M):
        assert M.data_size == (3, 5, 10)

    def test_data_access_float64(self, M):
        assert M[:].dtype == np.float64

    def test_movie_pixel_shape(self, M):
        assert M.pixel_shape == (5, 10)

    def test_movie_is_valid_pixel_coordinate(self, M):
        assert not M.is_valid_pixel_coordinate((0, -1))
        assert not M.is_valid_pixel_coordinate((4, 11))
        assert not M.is_valid_pixel_coordinate((4,))
        assert M.is_valid_pixel_coordinate((4, 9))

    def test_movie_extract_valid_pixels(self, M):
        assert M.extract_valid_pixels({(0, 0), (-1, 0), (4, 10)}) == {(0, 0)}

    def test_movie_get_item(self, M, movie_data):
        assert M[0, 0, 0] == 1.0
        np.testing.assert_allclose(M[2, :, :], movie_data[2, :, :])

    def test_movie_init_with_memmap(self, movie_data):
        # prepare memmapped file
        filename = os.path.join(TEST_DATA_DIR, "test_memdata.npy")
        mem_data = np.memmap(filename, dtype=np.uint16, mode="w+", shape=(3, 5, 10))
        mem_data[:] = movie_data[:]

        del mem_data

        # load memmapped file
        mem_data = np.memmap(filename, dtype=np.uint16, mode="r", shape=(3, 5, 10))
        movie = Movie("Memmap", mem_data)

        np.testing.assert_allclose(movie[:], movie_data)


class TestPatch:
    def test_patch_pixel_shape(self, simple_patch):
        assert simple_patch.pixel_shape == (7,)

    def test_patch_num_frames(self, simple_patch):
        assert simple_patch.num_frames == 3

    def test_patch_data(self, simple_patch, MM):
        np.testing.assert_equal(simple_patch[:], MM[:, 3:10])

    def test_patch_even_windowsize(self, MM):
        with pytest.raises(ValueError):
            Patch(MM, (5,), 6)

    def test_patch_enumerate_pixels(self, simple_patch, MM2):
        assert simple_patch.enumerate_pixels() == {
            (3,),
            (4,),
            (5,),
            (6,),
            (7,),
            (8,),
            (9,),
        }
        assert Patch(MM2, (5, 5), 3).enumerate_pixels() == {
            (4, 4),
            (4, 5),
            (4, 6),
            (5, 4),
            (5, 5),
            (5, 6),
            (6, 4),
            (6, 5),
            (6, 6),
        }

    @pytest.mark.parametrize(
        "movie_coordinate, patch_coordinate", [((6,), (3,)), ((3,), (0,))]
    )
    def test_patch_to_movie_coordinate(
        self, movie_coordinate, patch_coordinate, simple_patch, MM
    ):
        assert simple_patch.to_movie_coordinate(patch_coordinate) == movie_coordinate

    @pytest.mark.parametrize(
        "movie_coordinate, patch_coordinate", [((9,), (6,)), ((3,), (0,))]
    )
    def test_patch_to_patch_coordinate(
        self, movie_coordinate, patch_coordinate, simple_patch, MM
    ):
        assert simple_patch.to_patch_coordinate(movie_coordinate) == patch_coordinate


class TestSubsampler:
    def test_output_shape(self):
        assert Subsampler((800, 512, 512), 10).output_shape == (80, 512, 512)
        assert Subsampler((589, 128, 256), 10).output_shape == (59, 128, 256)
        assert Subsampler((589, 128, 256), 25).output_shape == (24, 128, 256)

    def test_buffer(self, subsampler):
        np.testing.assert_allclose(subsampler.buffer, np.zeros((10, 10, 10)))

        np.testing.assert_allclose(
            Subsampler((80, 512, 512), 10).buffer, np.zeros((8, 512, 512))
        )

    def test_buffer_last_frames(self, subsampler):
        subsampler = Subsampler((21, 10, 10), 2)
        subsampler.advance_buffer()

        assert subsampler.buffer.shape[0] == 1

    def test_add_frame(self, subsampler):
        subsampler.add_frame(np.ones((10, 10)))

        np.testing.assert_allclose(subsampler.buffer[0, :, :], np.ones((10, 10)))
        np.testing.assert_allclose(subsampler.buffer[1:, :, :], np.zeros((9, 10, 10)))

        subsampler.add_frame(np.ones((10, 10)) * 2)

        np.testing.assert_allclose(subsampler.buffer[0, :, :], 1.5 * np.ones((10, 10)))
        np.testing.assert_allclose(subsampler.buffer[1:, :, :], np.zeros((9, 10, 10)))

    def test_advance_to_next_frame_(self, subsampler):
        for i in range(10):
            subsampler.add_frame(np.ones((10, 10)))

        np.testing.assert_allclose(subsampler.buffer[0, :, :], np.ones((10, 10)))
        np.testing.assert_allclose(subsampler.buffer[1:, :, :], np.zeros((9, 10, 10)))

        subsampler.add_frame(np.ones((10, 10)) * 2)

        np.testing.assert_allclose(subsampler.buffer[0, :, :], np.ones((10, 10)))
        np.testing.assert_allclose(subsampler.buffer[1, :, :], 2 * np.ones((10, 10)))
        np.testing.assert_allclose(subsampler.buffer[2:, :, :], np.zeros((8, 10, 10)))

    def test_buffer_full(self, subsampler):
        for i in range(100):
            assert subsampler.buffer_full == False
            subsampler.add_frame(np.ones((10, 10)))
        assert subsampler.buffer_full == True

    def test_buffer_indices(self, subsampler):
        assert subsampler.buffer_indices == (0, 10)

    def test_advance_buffer(self, subsampler):
        for i in range(100):
            subsampler.add_frame(np.ones((10, 10)))

        subsampler.advance_buffer()
        np.testing.assert_allclose(subsampler.buffer, np.zeros((10, 10, 10)))
        assert subsampler.buffer_full == False
        assert subsampler.buffer_indices == (10, 20)

    def test_advance_buffer_resets_frame_count(self, subsampler):
        for i in range(100):
            subsampler.add_frame(np.ones((10, 10)))

        subsampler.advance_buffer()

        subsampler.add_frame(np.ones((10, 10)))

        np.testing.assert_allclose(subsampler.buffer[0, :, :], np.ones((10, 10)))

    def test_buffer_indices_last_frames(self):
        subsampler = Subsampler((21, 10, 10), 2)

        subsampler.advance_buffer()
        assert subsampler.buffer_indices == (10, 11)

    def test_prevent_buffer_overflow(self, subsampler):
        for i in range(100):
            subsampler.add_frame(np.ones((10, 10)))

        with pytest.raises(ValueError):
            subsampler.add_frame(np.ones((10, 10)))



# Copyright © 2017. Regents of the University of California (Regents). All Rights
# Reserved.
#
# Permission to use, copy, modify, and distribute this software and its documentation
# for educational, research, and not-for-profit purposes, without fee and without a
# signed licensing agreement, is hereby granted, provided that the above copyright
# notice, this paragraph and the following two paragraphs appear in all copies,
# modifications, and distributions. Contact The Office of Technology Licensing, UC
# Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620, (510) 643-7201,
# for commercial licensing opportunities. Created by Quico Spaen, Roberto Asín-Achá,
# and Dorit S. Hochbaum, Department of Industrial Engineering and Operations Research,
# University of California, Berkeley.
#
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
# INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE
# OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
# SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
# IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
"""Postprocesser component for selecting the best segmentation in HNCcorr."""

import numpy as np


class SizePostprocessor:
    """Selects the best segmentation based on the number of selected pixels.

    Discards all segmentations that contain more pixels than ``_max_size`` or less
    pixels then ``_min_size``. If no segmentations remains, no cell was found and
    ``None`` is returned. Otherwise the segmentation is returned that minimizes
    ``|sqrt(x) - sqrt(_pref_size)|`` where x is the number of pixels in the
    segmentation.

    Attributes:
        _min_size (int): Lower bound for the cell size in pixels.
        _max_size (int): Upper bound for the cell size in pixels.
        _pref_size (int): Preferred cell size in pixels.
    """

    def __init__(self, min_size, max_size, pref_size):
        """Initializes a SizePostprocessor object."""
        self._min_size = min_size
        self._max_size = max_size
        self._pref_size = pref_size

    def _filter(self, segmentations):
        """Returns a list of segmentations with size between min_size and max_size."""
        return [
            s
            for s in segmentations
            if self._min_size <= len(s.selection) <= self._max_size
        ]

    def select(self, segmentations):
        """Selects the best segmentation based on the number of selected pixels.

        See class description for details.

        Args:
            segmentations (List[Segmentation]): List of candidate segmentations.

        Returns:
            Segmentation or None: Best segmentation or None if all are discarded.
        """
        candidates = self._filter(segmentations)
        if not candidates:
            return None
        size = np.array([len(c.selection) for c in candidates])
        best_index = np.argmin(np.abs(np.sqrt(size) - np.sqrt(self._pref_size)))

        return candidates[best_index]
